/*
 * Copyright 2023 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef SDPMBODY_H
#define SDPMBODY_H

#include "sdpm.h"
#include "sdpmGroup.h"
#include "sdpmWake.h"
#include "sdpmMotion.h"
#include <map>

namespace sdpm
{

/**
 * @brief Manage a lifting body immersed in the fluid
 * @authors Adrien Crovato
 */
class SDPM_API Body : public Group
{
    // Geometry
    std::vector<Node *> _nodes;                      ///< nodes of the surface
    std::map<Node *, std::vector<Element *>> _neMap; ///< map between nodes and adjacent elements
    Wake *_wake;                                     ///< wake attached to the lifting body
    // Motion
    std::vector<Motion *> _motions;                      ///< body motions
    std::vector<std::vector<sdpmVector3cd>> _iVelocityS; ///< steady motion induced velocity
    std::vector<std::vector<sdpmVector3cd>> _iVelocityH; ///< harmonic motion induced velocity
    // Nodal aerodynamic load (normalized by dynamic pressure)
    std::vector<sdpmVector3d> _nLoads;                             ///< steady nodal aerodynamic load
    std::vector<std::vector<std::vector<sdpmVector3cd>>> _nLoads1; ///< unsteady first harmonic nodal aerodynamic load
    //  Coefficients
    sdpmDouble _cl;                             ///< steady lift coefficient
    sdpmDouble _cd;                             ///< steady drag coefficient
    sdpmDouble _cs;                             ///< steady side force coefficient
    sdpmDouble _cm;                             ///< steady pitch moment coefficient (positive nose-up)
    std::vector<std::vector<sdpmComplex>> _cl1; ///< unsteady first harmonic lift coefficient
    std::vector<std::vector<sdpmComplex>> _cd1; ///< unsteady first harmonic drag coefficient
    std::vector<std::vector<sdpmComplex>> _cs1; ///< unsteady first harmonic side force coefficient
    std::vector<std::vector<sdpmComplex>> _cm1; ///< unsteady first harmonic pitch moment coefficient (positive nose-up)

public:
    Body(Mesh &msh, std::string const &name, std::string const &teName, double shedLgth, size_t nShedDiv = 1, size_t nM = 0, size_t nF = 0, double ux = 1.);
    ~Body();

private:
    void getUniqueNodes();

public:
    std::vector<Node *> const &getNodes() const { return _nodes; }
    std::vector<sdpmVector3d> const &getLoads() const { return _nLoads; }
    std::vector<sdpmVector3cd> const &getUnsteadyLoads(size_t imd, size_t ifq) const { return _nLoads1[imd][ifq]; }
    sdpmDouble getLiftCoeff() const { return _cl; }
    sdpmDouble getDragCoeff() const { return _cd; }
    sdpmDouble getSideforceCoeff() const { return _cs; }
    sdpmDouble getMomentCoeff() const { return _cm; }
    sdpmComplex getUnsteadyLiftCoeff(size_t imd, size_t ifq) const { return _cl1[imd][ifq]; }
    sdpmComplex getUnsteadyDragCoeff(size_t imd, size_t ifq) const { return _cd1[imd][ifq]; }
    sdpmComplex getUnsteadySideforceCoeff(size_t imd, size_t ifq) const { return _cs1[imd][ifq]; }
    sdpmComplex getUnsteadyMomentCoeff(size_t imd, size_t ifq) const { return _cm1[imd][ifq]; }
    void addMotion(double ux, double amp, double xr);
#ifndef SWIG
    std::map<Node *, std::vector<Element *>> const &getMap() const
    {
        return _neMap;
    }
    Wake const &getWake() const { return *_wake; }
    size_t getNumMotions() const { return _motions.size(); }
    sdpmDouble getAmplitude(size_t imd) const { return _motions[imd]->getAmplitude(); }
    std::vector<sdpmVector3cd> const &getSteadyVelocity(size_t imd) const { return _iVelocityS[imd]; };
    std::vector<sdpmVector3cd> const &getHarmonicVelocity(size_t imd) const { return _iVelocityH[imd]; };
    void setLoads(std::vector<sdpmVector3d> const &nl) { _nLoads = nl; }
    void setUnsteadyLoads(size_t imd, size_t ifq, std::vector<sdpmVector3cd> const &nl) { _nLoads1[imd][ifq] = nl; }
    void setLiftCoeff(sdpmDouble cl) { _cl = cl; }
    void setDragCoeff(sdpmDouble cd) { _cd = cd; }
    void setSideforceCoeff(sdpmDouble cs) { _cs = cs; }
    void setMomentCoeff(sdpmDouble cm) { _cm = cm; }
    void setUnsteadyLiftCoeff(size_t imd, size_t ifq, sdpmComplex cl) { _cl1[imd][ifq] = cl; }
    void setUnsteadyDragCoeff(size_t imd, size_t ifq, sdpmComplex cd) { _cd1[imd][ifq] = cd; }
    void setUnsteadySideforceCoeff(size_t imd, size_t ifq, sdpmComplex cs) { _cs1[imd][ifq] = cs; }
    void setUnsteadyMomentCoeff(size_t imd, size_t ifq, sdpmComplex cm) { _cm1[imd][ifq] = cm; }
    void computeVelocity(size_t imd);
    virtual void write(std::ostream &out) const override;
#endif
};

} // namespace sdpm

#endif // SDPMBODY_H
