/*
 * Copyright 2023 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "sdpmBuilder.h"
#include "sdpmElement.h"
#include "sdpmNode.h"

using namespace sdpm;

/**
 * @brief Compute local frame of source panel
 */
sdpmMatrix3d Builder::computeTransfoMat(bool symy, Element const *ej)
{
    // Compute tangents and normal
    sdpmVector3d u = (ej->getCompressibleCoords()[0] - ej->getCompressibleCg()).normalized();
    sdpmVector3d v_ = ej->getCompressibleCoords()[1] - ej->getCompressibleCg();
    sdpmVector3d v = (v_ - (u.dot(v_)) * u).normalized();
    sdpmVector3d n = ej->getCompressibleNormal();

    // Assemble into transformation matrix
    sdpmMatrix3d g2l;
    g2l.row(0) = u;
    g2l.row(1) = v;
    g2l.row(2) = n;
    if (symy)
    {
        g2l.row(1) = -g2l.row(1);
        g2l.col(1) = -g2l.col(1);
    }
    return g2l;
}

/**
 * @brief Gather the coordinates of source panel
 */
sdpmMatrixXd Builder::gatherCoords(bool symy, Element const *ej)
{
    // Get compressible coordinates
    std::vector<sdpmVector3d> ccoords = ej->getCompressibleCoords();
    size_t nV = ccoords.size();

    // Get coordinates
    Eigen::MatrixXd coords(3, nV);
    for (size_t j = 0; j < nV; ++j)
        coords.col(j) = ccoords[nV - 1 - j];
    if (symy)
        coords.row(1) = -coords.row(1);
    return coords;
}

/**
 * @brief Compute vector joining body panel ej to body panel ei
 */
sdpmVector3d Builder::computeDistVec(bool symy, Element const *ei, Element const *ej)
{
    sdpmVector3d cgSrc = ej->getCompressibleCg();
    sdpmVector3d cgTgt = ei->getCompressibleCg();
    sdpmVector3d dist = cgTgt - cgSrc;
    if (symy)
        dist(1) = cgTgt(1) + cgSrc(1);
    return dist;
}

/**
 * @brief Compute doublet AIC and source AIC from body panel ej to body panel ei
 *
 * Katz and Plotkin, Low-speed aerodynamics, 2001
 * Hess and Smith, Calculation of potential flow about arbitrary bodies, 1967
 */
std::vector<sdpmDouble> Builder::computeAIC(bool symy, Element const *ei, Element const *ej, Eigen::MatrixXd const &sCoords, sdpmMatrix3d const &g2l)
{
    // Compute distance between panels and target panel coordinates in local axes
    sdpmVector3d distance = g2l * computeDistVec(symy, ei, ej);
    sdpmVector3d tCoords = g2l * ei->getCompressibleCg();

    // Compute coefficients
    int nV = static_cast<int>(ej->getNodes().size());
    Eigen::Vector4d d, r;
    for (int i = 1; i <= nV; ++i)
    {
        int j = i % nV + 1;
        d(i - 1) = sqrt((sCoords(0, j - 1) - sCoords(0, i - 1)) * (sCoords(0, j - 1) - sCoords(0, i - 1)) + (sCoords(1, j - 1) - sCoords(1, i - 1)) * (sCoords(1, j - 1) - sCoords(1, i - 1)));
        r(i - 1) = sqrt((tCoords(0) - sCoords(0, i - 1)) * (tCoords(0) - sCoords(0, i - 1)) + (tCoords(1) - sCoords(1, i - 1)) * (tCoords(1) - sCoords(1, i - 1)) + distance(2) * distance(2));
    }

    // Compute doublet AIC
    sdpmDouble mu = 0;
    if (ei->getId() == ej->getId() && !symy)
        mu = 0.5; // self-influence
    else
    {
        // additional coefficients
        Eigen::Vector4d e, h, F, G;
        for (int i = 1; i <= nV; ++i)
        {
            e(i - 1) = (tCoords(0) - sCoords(0, i - 1)) * (tCoords(0) - sCoords(0, i - 1)) + distance(2) * distance(2);
            h(i - 1) = (tCoords(0) - sCoords(0, i - 1)) * (tCoords(1) - sCoords(1, i - 1));
        }
        for (int i = 1; i <= nV; ++i)
        {
            int j = i % nV + 1;
            F(i - 1) = (sCoords(1, j - 1) - sCoords(1, i - 1)) * e(i - 1) - (sCoords(0, j - 1) - sCoords(0, i - 1)) * h(i - 1);
            G(i - 1) = (sCoords(1, j - 1) - sCoords(1, i - 1)) * e(j - 1) - (sCoords(0, j - 1) - sCoords(0, i - 1)) * h(j - 1);
        }
        // AIC
        for (int i = 1; i <= nV; ++i)
        {
            int j = i % nV + 1;
            mu += atan2(distance(2) * (sCoords(0, j - 1) - sCoords(0, i - 1)) * (F(i - 1) * r(j - 1) - G(i - 1) * r(i - 1)),
                        distance(2) * distance(2) * (sCoords(0, j - 1) - sCoords(0, i - 1)) * (sCoords(0, j - 1) - sCoords(0, i - 1)) * r(i - 1) * r(j - 1) + F(i - 1) * G(i - 1));
        }
        mu *= -1 / (4 * 3.14159);
    }

    // Compute source AIC
    sdpmDouble tau = 0;
    if (ei->getId() == ej->getId() && !symy)
    {
        for (int i = 1; i <= nV; ++i)
        {
            int j = i % nV + 1;
            tau += ((tCoords(0) - sCoords(0, i - 1)) * (sCoords(1, j - 1) - sCoords(1, i - 1)) - (tCoords(1) - sCoords(1, i - 1)) * (sCoords(0, j - 1) - sCoords(0, i - 1))) / d(i - 1) * log((r(i - 1) + r(j - 1) + d(i - 1)) / (r(i - 1) + r(j - 1) - d(i - 1)));
        }
        tau *= -1 / (4 * 3.14159); // self-influence
    }
    else
    {
        for (int i = 1; i <= nV; ++i)
        {
            int j = i % nV + 1;
            tau += ((tCoords(0) - sCoords(0, i - 1)) * (sCoords(1, j - 1) - sCoords(1, i - 1)) - (tCoords(1) - sCoords(1, i - 1)) * (sCoords(0, j - 1) - sCoords(0, i - 1))) / d(i - 1) * log((r(i - 1) + r(j - 1) + d(i - 1)) / (r(i - 1) + r(j - 1) - d(i - 1)));
        }
        tau = -1 / (4 * 3.14159) * tau - distance(2) * mu;
    }
    return std::vector<sdpmDouble>{mu, tau};
}
