/*
 * Copyright 2023 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "sdpmGmshExport.h"
#include "sdpmNode.h"
#include "sdpmElement.h"
#include "sdpmTag.h"
#include "sdpmResults.h"
#include <iostream>

using namespace sdpm;

GmshExport::GmshExport(Mesh &msh, bool b) : sdpmObject(), _msh(msh), _bin(b) {}
GmshExport::~GmshExport() { std::cout << "~sdpm::GmshExport()\n"; }

/**
 * @brief Save the mesh on disk
 */
void GmshExport::save(std::string const &sfx) const
{
    // write gmsh .msh file with stdio commands
    std::string wname = _msh.getName() + sfx + ".msh";
    std::cout << "writing file: " << wname << "... " << std::flush;
    FILE *file = fopen(wname.c_str(), "wt");
    if (!file)
        throw std::runtime_error("Error opening file (save)");

    // write format
    fprintf(file, "$MeshFormat\n");
    fprintf(file, "2.2 %d %d\n", 0, 8);
    fprintf(file, "$EndMeshFormat\n");

    // write physical names
    std::map<std::string, Tag *> tags = _msh.getTags();
    fprintf(file, "$PhysicalNames\n");
    fprintf(file, "%zu\n", static_cast<size_t>(tags.size()));
    for (auto &p : tags)
        fprintf(file, "%zu %zu \"%s\"\n", p.second->getDim(), p.second->getId(), p.second->getName().c_str());
    fprintf(file, "$EndPhysicalNames\n");

    // nodes
    std::vector<Node *> nods = _msh.getNodes();
    fprintf(file, "$Nodes\n");
    fprintf(file, "%zu\n", static_cast<size_t>(nods.size()));
    for (auto n : nods)
    {
        sdpmVector3d pos = n->getCoords();
        fprintf(file, "%zu %.16g %.16g %.16g\n", n->getId(), pos(0), pos(1), pos(2));
    }
    fprintf(file, "$EndNodes\n");

    // elements
    std::vector<Element *> elems = _msh.getElements();
    fprintf(file, "$Elements\n");
    fprintf(file, "%zu\n", static_cast<size_t>(elems.size()));
    for (auto e : elems)
    {
        fprintf(file, "%zu %zu %zu", e->getId(), static_cast<size_t>(e->getType()), static_cast<size_t>(2));
        // physical and elementary (equal to physical since not recorded) tags
        if (e->getTag())
            for (size_t i = 0; i < 2; ++i)
                fprintf(file, " %zu", e->getTag()->getId());
        else
            for (size_t i = 0; i < 2; ++i)
                fprintf(file, " 0");
        // nodes
        for (auto n : e->getNodes())
            fprintf(file, " %zu", n->getId());
        fprintf(file, "\n");
    }
    fprintf(file, "$EndElements\n");

    // close file
    fclose(file);
    std::cout << "done." << std::endl;
}

/**
 * @brief Save results (associated to a mesh) on disk
 */
void GmshExport::save(Results const &r, std::string const &sfx) const
{
    // write results to file
    std::string wname = _msh.getName() + sfx + ".pos";
    std::cout << "writing file: " << wname << "... " << std::flush;
    FILE *file;
    if (_bin)
        file = fopen(wname.c_str(), "wb");
    else
        file = fopen(wname.c_str(), "wt");
    if (!file)
        throw std::runtime_error("Error opening file (save)");

    // write format
    fprintf(file, "$MeshFormat\n");
    fprintf(file, "2.2 %d %d\n", _bin ? 1 : 0, 8);
    if (_bin)
    {
        int one = 1;
        fwrite(&one, sizeof(int), 1, file);
    }
    fprintf(file, "$EndMeshFormat\n");

    // get nodes and elements
    std::vector<Node *> nods = _msh.getNodes();
    std::vector<Element *> elems = _msh.getElements();

    // node data (scalar)
    for (auto &p : r.scalarsAtNodes)
    {
        fprintf(file, "$NodeData\n");
        fprintf(file, "%d\n", 1);                                 // nb of string tags
        fprintf(file, "\"%s\"\n", p.first.c_str());               // the name of the view
        fprintf(file, "%d\n", 1);                                 // nb of real tags (1)
        fprintf(file, "%.16g\n", r.time);                         //   1. the time value
        fprintf(file, "%d\n", 3);                                 // nb of integer tags (3)
        fprintf(file, "%d\n", r.nt);                              //   1. time step index
        fprintf(file, "%d\n", 1);                                 //   2. nb of field component (1, 3 or 9)
        fprintf(file, "%zu\n", static_cast<size_t>(nods.size())); //   3. nb of entities in the view
        for (size_t ii = 0; ii < nods.size(); ++ii)
        {
            if (_bin)
            {
                fwrite(&(nods[ii]->getIdRef()), sizeof(int), 1, file);
                fwrite(&((*(p.second))[ii]), sizeof(sdpmDouble), 1, file);
            }
            else
                fprintf(file, "%zu %.16g\n", nods[ii]->getId(), (*(p.second))[ii]);
        }
        fprintf(file, "$EndNodeData\n");
    }

    // element data (scalars)
    for (auto &p : r.scalarsAtElems)
    {
        fprintf(file, "$ElementData\n");
        fprintf(file, "%d\n", 1);                                  // nb of string tags
        fprintf(file, "\"%s\"\n", p.first.c_str());                // the name of the view
        fprintf(file, "%d\n", 1);                                  // nb of real tags (1)
        fprintf(file, "%.16g\n", r.time);                          //   1. the time value
        fprintf(file, "%d\n", 3);                                  // nb of integer tags (3)
        fprintf(file, "%d\n", r.nt);                               //   1. time step index
        fprintf(file, "%d\n", 1);                                  //   2. nb of field component (1, 3 or 9)
        fprintf(file, "%zu\n", static_cast<size_t>(elems.size())); //   3. nb of entities in the view
        for (size_t ii = 0; ii < elems.size(); ++ii)
        {
            if (_bin)
            {
                fwrite(&(elems[ii]->getIdRef()), sizeof(int), 1, file);
                fwrite(&(*(p.second))[ii], sizeof(sdpmDouble), 1, file);
            }
            else
                fprintf(file, "%zu %.16g\n", elems[ii]->getId(), (*(p.second))[ii]);
        }
        fprintf(file, "$EndElementData\n");
    }

    // node data (vectors)
    for (auto &p : r.vectorsAtNodes)
    {
        fprintf(file, "$NodeData\n");
        fprintf(file, "%d\n", 1);                                 // nb of string tags
        fprintf(file, "\"%s\"\n", p.first.c_str());               // the name of the view
        fprintf(file, "%d\n", 1);                                 // nb of real tags (1)
        fprintf(file, "%.16g\n", r.time);                         //   1. the time value
        fprintf(file, "%d\n", 3);                                 // nb of integer tags (3)
        fprintf(file, "%d\n", r.nt);                              //   1. time step index
        fprintf(file, "%d\n", 3);                                 //   2. nb of field component (1, 3 or 9)
        fprintf(file, "%zu\n", static_cast<size_t>(nods.size())); //   3. nb of entities in the view
        for (size_t ii = 0; ii < nods.size(); ++ii)
        {
            sdpmVector3d const &v = (*(p.second))[ii];
            if (_bin)
            {
                fwrite(&(nods[ii]->getIdRef()), sizeof(int), 1, file);
                fwrite(&(v(0)), sizeof(sdpmDouble), 3, file);
            }
            else
                fprintf(file, "%zu %.16g %.16g %.16g\n", nods[ii]->getId(), v(0), v(1), v(2));
        }
        fprintf(file, "$EndNodeData\n");
    }

    // element data (vectors)
    for (auto &p : r.vectorsAtElems)
    {
        fprintf(file, "$ElementData\n");
        fprintf(file, "%d\n", 1);                                  // nb of string tags
        fprintf(file, "\"%s\"\n", p.first.c_str());                // the name of the view
        fprintf(file, "%d\n", 1);                                  // nb of real tags (1)
        fprintf(file, "%.16g\n", r.time);                          //   1. the time value
        fprintf(file, "%d\n", 3);                                  // nb of integer tags (3)
        fprintf(file, "%d\n", r.nt);                               //   1. time step index
        fprintf(file, "%d\n", 3);                                  //   2. nb of field component (1, 3 or 9)
        fprintf(file, "%zu\n", static_cast<size_t>(elems.size())); //   3. nb of entities in the view
        for (size_t ii = 0; ii < elems.size(); ++ii)
        {
            sdpmVector3d const &v = (*(p.second))[ii];
            if (_bin)
            {
                fwrite(&(elems[ii]->getIdRef()), sizeof(int), 1, file);
                fwrite(&(v(0)), sizeof(sdpmDouble), 3, file);
            }
            else
                fprintf(file, "%zu %.16g %.16g %.16g\n", elems[ii]->getId(), v(0), v(1), v(2));
        }
        fprintf(file, "$EndElementData\n");
    }

    // close file
    fclose(file);
    std::cout << "done." << std::endl;
}

void GmshExport::write(std::ostream &out) const
{
    out << "GmshExport on " << _msh << std::endl;
}
