/*
 * Copyright 2023 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "sdpmProblem.h"
#include "sdpmBody.h"
#include "sdpmElement.h"
#include "sdpmTag.h"
#include <iostream>
using namespace sdpm;

Problem::Problem(Mesh &msh, size_t nmods, std::vector<double> const &fs) : _msh(msh),
                                                                           _aoa(0.), _aos(0.), _mach(0.), _beta(1.0), _uinf(sdpmVector3d::Zero()),
                                                                           _dDir(sdpmVector3d::Zero()), _sDir(sdpmVector3d::Zero()), _lDir(sdpmVector3d::Zero()),
                                                                           _sRef(0.), _cRef(0.), _xRef(sdpmVector3d::Zero()), _ySym(false),
                                                                           _nModes(nmods), _omega(fs),
                                                                           _hasAero(false), _hasGeo(false)
{
}

Problem::~Problem()
{
    std::cout << "~sdpm::Problem()\n";
}

/**
 * @brief Add a non-lifting body
 */
void Problem::addBody(Body &b)
{
    _bodies.push_back(&b);
}

/**
 * @brief Update the freestream aerodynamic values
 */
void Problem::updateFreestream(double aoa, double aos, double mch, double ui)
{
    // Angles
    _aoa = aoa;
    _aos = aos;
    // Compressiblity
    _mach = mch;
    _beta = std::sqrt(1 - _mach * _mach);
    // Velocity and directions
    _uinf = ui * sdpmVector3d(cos(_aoa) * cos(_aos), sin(_aos), sin(_aoa) * cos(_aos));
    _dDir = sdpmVector3d(cos(_aoa) * cos(_aos), sin(_aos), sin(_aoa) * cos(_aos));
    _sDir = sdpmVector3d(-cos(_aoa) * sin(_aos), cos(_aos), -sin(_aoa) * sin(_aos));
    _lDir = sdpmVector3d(-sin(_aoa), 0., cos(_aoa));
    // Flag
    _hasAero = true;
}

/**
 * @brief Set the reference geometric values
 */
void Problem::setGeometry(double sref, double cref, double xref, double yref, double zref, bool ysym)
{
    // Reference values
    _sRef = sref;
    _cRef = cref;
    _xRef = sdpmVector3d(xref, yref, zref);
    _ySym = ysym;
    // Flag
    _hasGeo = true;
}

/**
 * @brief Update the elements values
 */
void Problem::updateElements()
{
    for (auto b : _bodies)
    {
        for (auto e : b->getElements())
            e->update(_beta);
        for (auto e : b->getWake().getElements())
            e->update(_beta);
    }
}

/**
 * @brief Check that Problem is not empty and that elements are supported
 */
void Problem::check() const
{
    // Sanity checks
    if (!_hasAero)
        throw std::runtime_error("sdpm::Problem: aerodynamic parameters have not been provided!\n");
    if (!_hasGeo)
        throw std::runtime_error("sdpm::Problem: geometric parameters have not been provided!\n");
    if (_bodies.empty())
        throw std::runtime_error("sdpm::Problem: no lifting bodies provided!\n");
    for (auto b : _bodies)
        if (b->getNumMotions() != _nModes)
            throw std::runtime_error("sdpm::Problem: Body and Problem do not have the same number of modes!\n");

    // Surface elements
    for (auto b : _bodies)
        for (auto e : b->getElements())
            if (e->getType() != ElType::QUAD4)
            {
                std::stringstream err;
                err << "SDPM is only implemented for surface elements of type "
                    << ElType::QUAD4 << " (" << e->getType() << " was given)!\n";
                throw std::runtime_error(err.str());
            }
}

void Problem::write(std::ostream &out) const
{
    out << "sdpm::Problem definition"
        << "\n\tAngle of attack: " << _aoa
        << "\n\tAngle of sideslip: " << _aos
        << "\n\tMach number: " << _mach
        << "\n\tReference area: " << _sRef
        << "\n\tReference length: " << _cRef
        << "\n\tReference point: [" << _xRef.transpose() << "]"
        << std::endl;
    out << "with";
    for (auto b : _bodies)
        out << "\t" << *b;
}
