/*
 * Copyright 2023 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef SDPMSOLVER_H
#define SDPMSOLVER_H

#include "sdpm.h"
#include "sdpmObject.h"
#include <vector>
#include <map>

namespace sdpm
{

/**
 * @brief Source-doublet panel solver
 * @authors Adrien Crovato
 */
class SDPM_API Solver : public sdpmObject
{
    // Problem
    Problem &_pbl; ///< problem definition
    // AICs
    size_t _nP;                     ///< number of body panels
    std::map<Element *, int> _rows; ///< map linking an element to a matrix cell
    sdpmMatrixXd _A0;               ///< zero-frequency body-body doublet matrix
    sdpmMatrixXd _B0;               ///< zero-frequency body-body source matrix
    std::vector<sdpmMatrixXcd> _A;  ///< unsteady body-body doublet matrix
    std::vector<sdpmMatrixXcd> _B;  ///< unsteady body-body source matrix
    // Solution
    std::vector<sdpmDouble> _phi;                             ///< perturbation potential
    std::vector<sdpmDouble> _tau;                             ///< source strength
    std::vector<sdpmDouble> _mu;                              ///< doublet strength
    std::vector<sdpmVector3d> _u;                             ///< steady velocity
    std::vector<sdpmDouble> _cp;                              ///< steady pressure coefficient
    std::vector<std::vector<std::vector<sdpmVector3cd>>> _u1; ///< unsteady first harmonic velocity
    std::vector<std::vector<std::vector<sdpmComplex>>> _cp1;  ///< unsteady first harmonic pressure coefficient
    // Aerodynamic loads
    sdpmDouble _cl;                             ///< steady lift coefficient
    sdpmDouble _cd;                             ///< steady drag coefficient
    sdpmDouble _cs;                             ///< steady sideforce coefficient
    sdpmDouble _cm;                             ///< steady pitch moment coefficient (positive nose-up)
    std::vector<std::vector<sdpmComplex>> _cl1; ///< unsteady first harmonic lift coefficient
    std::vector<std::vector<sdpmComplex>> _cd1; ///< unsteady first harmonic drag coefficient
    std::vector<std::vector<sdpmComplex>> _cs1; ///< unsteady first harmonic side force coefficient
    std::vector<std::vector<sdpmComplex>> _cm1; ///< unsteady first harmonic pitch moment coefficient (positive nose-up)

public:
    Solver(Problem &pbl);
    ~Solver();

    std::vector<sdpmDouble> const &getPressure() const { return _cp; }
    std::vector<sdpmComplex> const &getUnsteadyPressure(size_t imd, size_t ifq) const { return _cp1[imd][ifq]; }
    sdpmDouble getLiftCoeff() const { return _cl; }
    sdpmDouble getDragCoeff() const { return _cd; }
    sdpmDouble getSideforceCoeff() const { return _cs; }
    sdpmDouble getMomentCoeff() const { return _cm; }
    sdpmComplex getUnsteadyLiftCoeff(size_t imd, size_t ifq) const { return _cl1[imd][ifq]; }
    sdpmComplex getUnsteadyDragCoeff(size_t imd, size_t ifq) const { return _cd1[imd][ifq]; }
    sdpmComplex getUnsteadySideforceCoeff(size_t imd, size_t ifq) const { return _cs1[imd][ifq]; }
    sdpmComplex getUnsteadyMomentCoeff(size_t imd, size_t ifq) const { return _cm1[imd][ifq]; }

    void update();
    void run();
    void runUnsteady();
    void save(GmshExport const &writer, std::string const &suffix = "");

#ifndef SWIG
    virtual void write(std::ostream &out) const override;
#endif

private:
    void post(sdpmVectorXd const &etau, sdpmVectorXd const &emu);
    void postUnsteady(size_t imd, size_t ifq, sdpmVectorXcd const &etau, sdpmVectorXcd const &emu);
};

} // namespace sdpm
#endif // SDPMSOLVER_H
