# -*- coding: utf-8 -*-

# Copyright 2023 University of Liège
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# LANN wing
# Adrien Crovato

import sdpm
from sdpm.gmsh import MeshLoader
from sdpm.utils import build_fpath
from sdpm.testing import *
import math
import cmath

def main():
    # geometry parameters
    wnc = 20 # number of chordwise panels
    pars = {'nC': wnc, 'bC': 0.25, 'rS': 1}
    # flow and time parameters
    minf = 0.621 # freestream Mach number
    aoa = 0.59 * math.pi / 180
    omega = 2 * math.pi * 24 # circular frequency
    amp = 0.5 * 0.25 * math.pi / 180 # semi pitching amplitude
    uinf = 0.5 * 0.361 * omega / 0.133; # freestream velocity

    # start timer
    tms = sdpm.Timers()
    tms['total'].start()
    # mesh
    tms['msh'].start()
    msh = MeshLoader(build_fpath('models/lann.geo', __file__, 1)).run(pars)
    tms['msh'].stop()
    # problem
    tms['pbl'].start()
    pbl = sdpm.Problem(msh, 1, [omega])
    pbl.setGeometry(0.253, 0.361, 0.224, 0., 0., True)
    pbl.updateFreestream(aoa, 0., minf, uinf)
    wing = sdpm.Body(msh, 'wing', 'te', 0.361, wnc, 1, 1, uinf*math.cos(aoa))
    wing.addMotion(uinf*math.cos(aoa), amp, 0.224/math.sqrt(1-minf*minf))
    pbl.addBody(wing)
    tms['pbl'].stop()
    ## solver
    tms['sol'].start()
    sol = sdpm.Solver(pbl)
    sol.update()
    sol.run()
    sol.runUnsteady()
    sol.save(sdpm.GmshExport(msh))
    tms['sol'].stop()
    # end timer
    tms['total'].stop()
    print(tms)

    # test
    tests = CTests()
    tests.add(CTest('CL0', sol.getLiftCoeff(), 0.264, 5e-2)) # MATLAB: 0.27
    tests.add(CTest('CD0', sol.getDragCoeff(),  0.0020, 5e-2)) # MATLAB: 0.0031
    tests.add(CTest('CS0', sol.getSideforceCoeff(), 0.008, 5e-2))
    tests.add(CTest('CM0', sol.getMomentCoeff(), -0.079, 5e-2))
    tests.add(CTest('Abs(CL1)', abs(sol.getUnsteadyLiftCoeff(0, 0)) / amp / math.pi, 1.61, 5e-2)) # MATLAB: 1.69
    tests.add(CTest('Ang(CL1)', cmath.phase(sol.getUnsteadyLiftCoeff(0, 0)) * 180 / math.pi, 0.46, 5e-2)) # MATLAB: 0.64
    tests.add(CTest('Abs(CD1)', abs(sol.getUnsteadyDragCoeff(0, 0)) / amp / math.pi, 0.036, 5e-2)) # MATLAB: 0.037
    tests.add(CTest('Ang(CD1)', cmath.phase(sol.getUnsteadyDragCoeff(0, 0)) * 180 / math.pi, 16.4, 5e-2)) # MATLAB: 19.9
    tests.add(CTest('Abs(CS1)', abs(sol.getUnsteadySideforceCoeff(0, 0)) / amp / math.pi, 0.037, 5e-2))
    tests.add(CTest('Ang(CS1)', cmath.phase(sol.getUnsteadySideforceCoeff(0, 0)) * 180 / math.pi,  -10.0, 5e-2))
    tests.add(CTest('Abs(CM1)', abs(sol.getUnsteadyMomentCoeff(0, 0)) / amp / math.pi, 0.33, 5e-2))
    tests.add(CTest('Ang(CM1)', cmath.phase(sol.getUnsteadyMomentCoeff(0, 0)) * 180 / math.pi, -170.5, 5e-2))
    tests.run()

    # eof
    print('')

if __name__ == '__main__':
    main()
