# -*- coding: utf-8 -*-

# Copyright 2020 University of Liège
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Onera M6 wing
# Adrien Crovato

import sdpm
from sdpm.gmsh import MeshLoader
from sdpm.utils import build_fpath
from sdpm.testing import *
import math

def main():
    # geometry parameters
    pars = {'nC': 100, 'nS': 10, 'bC': 0.25, 'pS': 1.0}
    # flow parameters
    aoa = 3.06 * math.pi / 180
    minf = 0.839

    # start timer
    tms = sdpm.Timers()
    tms['total'].start()
    # mesh
    tms['msh'].start()
    msh = MeshLoader(build_fpath('models/oneraM6.geo', __file__, 1)).run(pars)
    tms['msh'].stop()
    # problem
    tms['pbl'].start()
    pbl = sdpm.Problem(msh)
    pbl.setGeometry(0.7528, 0.64, 0., 0., 0., True)
    pbl.updateFreestream(aoa, 0., minf, 1.)
    wing = sdpm.Body(msh, 'wing', 'te', 0.805)
    pbl.addBody(wing)
    tms['pbl'].stop()
    ## solver
    tms['sol'].start()
    sol = sdpm.Solver(pbl)
    sol.update()
    sol.run()
    sol.save(sdpm.GmshExport(msh))
    tms['sol'].stop()
    # end timer
    tms['total'].stop()
    print(tms)

    # test
    tests = CTests()
    tests.add(CTest('CL', sol.getLiftCoeff(), 0.255, 5e-2)) # panair: 0.255
    tests.add(CTest('CD', sol.getDragCoeff(), 0.005, 5e-2)) # panair: 0.005
    tests.add(CTest('CS', sol.getSideforceCoeff(), 0.011, 5e-2)) # panair: 0.011
    tests.add(CTest('CM', sol.getMomentCoeff(), -0.189, 5e-2)) # panair: -0.189
    tests.run()

    # eof
    print('')

if __name__ == '__main__':
    main()
