# -*- coding: utf-8 -*-

# Copyright 2023 University of Liège
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

def parse_args():
    """Parses command line arguments
    """
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument('-v', '--verb', help='increase output verbosity', action='count', default=0)
    parser.add_argument('--nogui', help='disable any graphical output', action='store_true')
    parser.add_argument('--clean', help='clean workspace', action='store_true')
    parser.add_argument('file', help='python files')
    args = parser.parse_args()
    return args

def build_fpath(name, caller, recurse_lvl=0):
    """Build absolute file path from a target and a caller

    Parameters
    ----------
    name : string
        name of the target file
    caller : string
        name of the file from which the target should be located
    recurse_lvl : int
        number of times to recurse to parent directory
    """
    import os.path
    recurse = ''
    for _ in range(recurse_lvl):
        recurse = os.path.join(recurse, '..')
    return os.path.abspath(os.path.join(os.path.dirname(caller), recurse, name))

def sort(conn, pts_vals):
    """Sort value array against line connectivity list

    Parameters
    ----------
    connct : 2d numpy array
        connectivity list (n_cells X n_point_by_cell)
    pts_vals : 2d numpy array
        points coordinates and values
    """
    import numpy as np
    # sort id vector
    conn = conn[conn[:,0].argsort(), :]
    # deep copy of data
    _copy = np.zeros((pts_vals.shape[0], pts_vals.shape[1]))
    _copy[:] = pts_vals
    # sort data against id
    _max = np.argmax(pts_vals[:, 0])
    pts_vals[0,:] = _copy[_max, :]
    nextId = conn[np.where(conn[:, 0]==_max)[0][0], 0]
    for i in range(1, conn.shape[0]):
        pts_vals[i,:] = _copy[conn[nextId,1], :]
        nextId = conn[nextId, 1]

def create_slices(body, sol, fname, ys):
    """Create slices along the span and extract the chordwise pressure coefficient distribution

    Parameters
    ----------
    body : sdpmBody
        body data structure containing grid points and cells
    sol : dict(string, array)
        name and solution data at cells
    fname : string
        name of the .vtu file (without extension)
    ys : array
        y-coordinates defining cutplanes
    """
    from sdpm.vtk_utils import write_grid, cut_grid, extract_data
    import numpy as np
    # Create grid and write to file
    grid = write_grid(fname, body, sol)
    # Extract data at along spanwise sections
    snames = list(sol.keys())
    for i in range(len(ys)):
        # get data on cut section
        coords, connct, cvalues = extract_data(cut_grid(grid, [0., ys[i], 0.]), snames)
        # normalize x-coordinate by local chord
        xc = np.zeros((coords.shape[0], 1))
        xc[:,0] = (coords[:,0] - min(coords[:,0])) / (max(coords[:,0]) - min(coords[:,0]))
        # average value at points from values at cell center
        nvalues = np.zeros((coords.shape[0], len(sol.keys())))
        for j in range(connct.shape[0]):
            pids = connct[j]
            for k in range(len(pids)):
                for l in range(len(sol.keys())):
                    nvalues[connct[j][k], l] += cvalues[snames[l]][j] / len(pids)
        # create dataset holding x-coordinates and values and sort it
        xc_vals = np.hstack((xc, nvalues))
        sort(connct, xc_vals)
        xc_vals = np.vstack((xc_vals, xc_vals[0,:]))
        # write to file
        hdr = 'x/c'
        for name in snames:
            hdr += f', {name}'
        hdr += f'\ny = {ys[i]}'
        np.savetxt(fname+f'_slice_{i}.dat', xc_vals, fmt='%1.5e', delimiter=',', header=hdr)
